def TSM3_FKIKSwitch():
    import pymel.core as pa
    import maya.mel as mel
    import re
    from maya.OpenMaya import MGlobal
    import pymel.core.datatypes as dt    
# Create a null as a sibling of the target control,
# constrain it to the source of the numbers
#copy over the values
    def usePV():
        return pa.menuItem('TSM3ControlsFKIKPoleVector', query=True, checkBox=True)

    def saferHasAttr(node, attr):
        try:
            return node.hasAttr(attr)
        except KeyError:
            return False
    
    def bodyPartPMA(bodyPartId):
        return [node for node in pa.ls(type='plusMinusAverage') if saferHasAttr(node, 'TSM3BodyPartId') and node.TSM3BodyPartId.get() == bodyPartId]

    def bodyPartTransforms(bodyPartId):
        return [node for node in pa.ls(type='transform') if saferHasAttr(node, 'TSM3BodyPartId') and node.TSM3BodyPartId.get() == bodyPartId]
    
    def nodeWithFKIKTag(nodes, tag):
        for node in nodes:
            if node.hasAttr('IKFKTag') and node.IKFKTag.get() == tag:
                return node
        return None
        
    def nodeWithId(nodes, id):
        for node in nodes:
            if node.hasAttr('TSM3NodeId') and node.TSM3NodeId.get() == id:
                return node
        return None

    def transferRotation(source, target):
        null = pa.duplicate(target, po=True)[0]
        null.rotateX.set(lock=False)
        null.rotateY.set(lock=False)
        null.rotateZ.set(lock=False)
        pa.orientConstraint(source, null)
        target.setRotation(null.getRotation())
        pa.delete(null);
        
    def transferPosition(source, target):
        null = pa.group(em=True, parent = target.getParent())
        pa.parentConstraint(source, null)
        target.setTranslation(null.getTranslation())
        pa.delete(null)

    def transferPositionAndRotation(source, target):
        transferRotation(source, target)
        transferPosition(source, target)

    def transferToPV(poleVectorIK, arrowIK):
        pvWorldSpace = poleVectorIK.getTranslation(space='world')
        arrowIK.rotate.set(0,0,0)
        poleVectorIK.setTranslation(pvWorldSpace,space='world')

    def armIKToFKTransfer(bodyPartId, bodyPartNodes):
        outNodes=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperArmOut', 'lowerArmOut', 'handOut']]
        inFKNodes=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperArmInFK', 'lowerArmInFK', 'handInFK']]

        stretchNode=nodeWithId(bodyPartNodes, 'stretchValues')
        inFKNodes[0].stretch.set(stretchNode.UpperArmStretch.get())
        inFKNodes[1].stretch.set(stretchNode.LowerArmStretch.get())
            
        [transferRotation(source, target) for source, target in zip(outNodes, inFKNodes)] 
        
    def armFKToIKTransfer(bodyPartId, bodyPartNodes):
        handOut,handInIK=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['handOut', 'handInIK']]

        stretchNode=nodeWithId(bodyPartNodes, 'stretchValues')
        handInIK.UpperArmStretch.set(stretchNode.UpperArmStretch.get())
        handInIK.LowerArmStretch.set(stretchNode.LowerArmStretch.get())

        transferRotation(handOut, handInIK)
        transferPosition(handOut, handInIK)
        
        armFKToIKTransferRotate(bodyPartId,bodyPartNodes)

        if usePV(): armFKToIKTransferPV(bodyPartId,bodyPartNodes)

    def armFKToIKTransferPV(bodyPartId, bodyPartNodes):
        lowerArm, poleVectorIK, arrowIK = [nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['lowerArmOut', 'poleVectorArmCtrlIK','upperArmInIK']]

        transferToPV(poleVectorIK, arrowIK)

    def armFKToIKTransferRotate(bodyPartId, bodyPartNodes):
        upperArmOut, upperArmInIK, poleVectorIK =[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperArmOut','upperArmInIK','poleVectorArmCtrlIK']]

        transferRotation(upperArmOut, upperArmInIK)
        poleVectorIK.translate.set(0,0,0)

    def legIKToFKTransfer(bodyPartId, bodyPartNodes):
        outNodes=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperLegOut', 'lowerLegOut', 'footOut', 'toeOut']]
        inFKNodes=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperLegInFK', 'lowerLegInFK', 'footInFK', 'toeInFK']]

        stretchNode=nodeWithId(bodyPartNodes, 'stretchValues')
        inFKNodes[0].stretch.set(stretchNode.UpperLegStretch.get())
        inFKNodes[1].stretch.set(stretchNode.LowerLegStretch.get())
        inFKNodes[2].stretch.set(stretchNode.FootStretch.get())
            
        [transferRotation(source, target) for source, target in zip(outNodes, inFKNodes)] 

    def legFKToIKTransfer(bodyPartId, bodyPartNodes):
        #Leg FK to IK can either go to the PV control or the rotation control, but the first part is the same.  
        print 'even in transfer'
        footIKOut, toeIKOut=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['footIKOut', 'toeIKOut']]
        footInIK, toeInIK=[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['footInIK', 'toeInIK']]

        stretchNode=nodeWithId(bodyPartNodes, 'stretchValues')
        footInIK.UpperLegStretch.set(stretchNode.UpperLegStretch.get())
        footInIK.LowerLegStretch.set(stretchNode.LowerLegStretch.get())
        footInIK.FootStretch.set(stretchNode.FootStretch.get())
    
        transferRotation(footIKOut, footInIK)
        transferPosition(footIKOut, footInIK)
        transferRotation(toeIKOut, toeInIK)

        legFKToIKTransferRotate(bodyPartId, bodyPartNodes)
        
        if usePV(): legFKToIKTransferPV(bodyPartId, bodyPartNodes)
        
    def legFKToIKTransferPV(bodyPartId, bodyPartNodes):
        print 'transfer pv'
        lowerLeg, poleVectorIK, arrowIK =[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['lowerLegOut','poleVectorLegCtrlIK','upperLegInIK']]

        transferToPV(poleVectorIK, arrowIK)

    def legFKToIKTransferRotate(bodyPartId, bodyPartNodes):
        upperLegOut, upperLegInIK, poleVectorIK =[nodeWithFKIKTag(bodyPartNodes, tag) for tag in ['upperLegOut', 'upperLegInIK','poleVectorLegCtrlIK']]
        transferRotation(upperLegOut, upperLegInIK)
        poleVectorIK.translate.set(0,0,0)

    def numberOfTailControls(bodyPartNodes):
        return len([node for node in bodyPartNodes if re.match(r'TailFKControl\d+$', node.TSM3NodeId.get())])
        
    def tailIKToFKTransfer(bodyPartId, bodyPartNodes):
        for suffix in range(1, numberOfTailControls(bodyPartNodes)+1):
            source=nodeWithId(bodyPartNodes, 'TailIKControl' + str(suffix))
            target=nodeWithId(bodyPartNodes, 'TailFKControl' + str(suffix))
            
            transferRotation(source, target)
            transferPosition(source, target)
            
    def tailFKToIKTransfer(bodyPartId, bodyPartNodes):
        for suffix in range(1, numberOfTailControls(bodyPartNodes)+1):
            source=nodeWithId(bodyPartNodes, 'TailFKControl' + str(suffix))
            target=nodeWithId(bodyPartNodes, 'TailIKControl' + str(suffix))
            
            transferRotation(source, target)
            transferPosition(source, target)
      
    def spineIKToFKTransfer(bodyPartId, bodyPartNodes):
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'hipsIKControl'),
                                    nodeWithId(bodyPartNodes, 'lowerSpineControl'))
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'middleSpineIKControl'),
                                    nodeWithId(bodyPartNodes, 'middleSpineControl'))
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'torsoIKControl'),
                                    nodeWithId(bodyPartNodes, 'torsoControl'))
        nodeWithId(bodyPartNodes, 'hipsControl').t.set(0,0,0)
        nodeWithId(bodyPartNodes, 'hipsControl').r.set(0,0,0)

    def spineFKToIKTransfer(bodyPartId, bodyPartNodes):
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'torsoControl'),
                                    nodeWithId(bodyPartNodes, 'torsoIKControl'))
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'hipsControl'),
                                    nodeWithId(bodyPartNodes, 'hipsIKControl'))
        transferPositionAndRotation(nodeWithId(bodyPartNodes, 'middleSpineControl'),
                                    nodeWithId(bodyPartNodes, 'middleSpineIKControl'))

    def bodyPartFKIKSwitch(bodyPartId):
        bodyPartNodes=bodyPartTransforms(bodyPartId)
        connectsToUpperBodyNode=nodeWithId(bodyPartNodes, 'AttributesConnectToUpperBody')
        bodyPartType=connectsToUpperBodyNode.bodyPartType.get()
            
        switchFunctions={'arm': {'FKtoIK': armFKToIKTransfer, 'IKtoFK': armIKToFKTransfer},
                         'leg': {'FKtoIK': legFKToIKTransfer, 'IKtoFK': legIKToFKTransfer},
                         'tail': {'FKtoIK': tailFKToIKTransfer, 'IKtoFK': tailIKToFKTransfer},
                         'spine': {'FKtoIK': spineFKToIKTransfer, 'IKtoFK': spineIKToFKTransfer}}
        
        if bodyPartType not in switchFunctions:
            MGlobal.displayWarning('Cannot switch body part of type "' + bodyPartType + '"')
            return
        
        FKIKSwitchAttr=connectsToUpperBodyNode.FKIKSwitch.inputs(p=True)[0]
        FKIKValue=FKIKSwitchAttr.get()
        
        if FKIKValue != 0.0 and FKIKValue != 1.0:
            MGlobal.displayWarning('Cannot switch ' + FKIKSwitchAttr.name() + '. It must be either 0 or 1.')
            return
        
        if FKIKValue == 0.0:
            switchType='FKtoIK'
        else:
            switchType='IKtoFK'
            
        switchFunctions[bodyPartType][switchType](bodyPartId, bodyPartNodes)
        FKIKSwitchAttr.set(1.0-FKIKValue)
    
    def selectedBodyParts():
        return set([node.TSM3BodyPartId.get() for node in pa.ls(sl=True, type='transform') if node.hasAttr('TSM3BodyPartId')])
    
    for bodyPartId in selectedBodyParts():
        bodyPartFKIKSwitch(bodyPartId)

TSM3_FKIKSwitch()
