def TSM3_MocapBakeControls():
    import maya.OpenMaya as om
    import pymel.core as pa
    import traceback

    def associatedRigNode(node):
        """Find the "Rig" transform node based on a part of the rig.""" 
        return [parent for parent in node.getAllParents() if parent.hasAttr('TSM3NodeId') and parent.TSM3NodeId.get() == 'TSM3_Rig'][0]

    def upperBodyFromRigNode(rigNode):
        """Returns upper body node, given the "Rig" transform"""
        upperBodyList = [joint for joint in pa.listRelatives(rigNode, ad = True, typ = "transform")
            if joint.hasAttr("TSM3NodeId") and joint.TSM3NodeId.get() == "TSM3_upper_body"]
        if len(upperBodyList) == 1:
            return upperBodyList[0]
        else:
            return False

    def bakeMocapSkeleton(rigNodeChildren, upperBodyControl):
        """Bake the current motion on the mocap skeleton."""
        controls = [node for node in rigNodeChildren if node.hasAttr("TSM3NodeId")
                    and node.hasAttr('TSM3Control')]
        #controls = controls + [upperBodyControl]

        timeParam = (pa.playbackOptions(animationStartTime=True, q=True), pa.playbackOptions(animationEndTime=True, q=True))

        pa.bakeResults(controls, simulation=True, 
                       sampleBy=1, time=timeParam,
                       disableImplicitControl=True, preserveOutsideKeys=True, hierarchy=False,
                       sparseAnimCurveBake=False, removeBakedAttributeFromLayer=False, 
                       bakeOnOverrideLayer=False, controlPoints=False, shape=False)

    def deleteMocapSkeleton(rigNodeChildren):
        """Delete the mocap skeleton, its constraints and connection nodes."""
        [pa.delete(node) for node in rigNodeChildren if node.hasAttr("TSM3DupeNodeId")
                        if node.TSM3DupeNodeId.get().endswith("MocapConstraint")
                        or node.TSM3DupeNodeId.get() == "TSM3WorldJoint"]
    def rigNodeHasDupeSkeleton(rigNode):
        dupeJointList = [joint for joint in pa.listRelatives(rigNode, ad = True)
                    if joint.hasAttr("TSM3DupeBodyPartId")]
        return len(dupeJointList) is not 0

    def verifySelection():
        """Checks to make sure a TSM3 joint or control is found"""
        try:
            rigNode = associatedRigNode(pa.ls(sl=True)[0])
            if not rigNodeHasDupeSkeleton(rigNode):
                    raise Exception("Selection Invalid: No duplicate skeleton exists for the selected TSM3 rig.")
        except IndexError:
            raise Exception("Selection invalid:To bake the animation of the motion capture skeleton, please select part of the TSM3 rig.")


    def bakeAndDeleteMocapSkeleton():
        rigNode = associatedRigNode(pa.ls(sl=True)[0])
        upperBodyControl = upperBodyFromRigNode(rigNode)
        rigNodeChildren = rigNode.listRelatives(ad=True)
        bakeMocapSkeleton(rigNodeChildren, upperBodyControl)
        deleteMocapSkeleton(rigNodeChildren)

    verifySelection()
    bakeAndDeleteMocapSkeleton()
