def TSM3_MocapSkeleton():
	import pymel.core as pa
	import maya.OpenMaya as om
	import traceback
	import maya.cmds as cmds

	mocapDict = {
			("head","head"):[("Control","parent")],
			("head","neck"):[("neckControl","orient")],
			("spine","hiResSpine1"):[("hiResSpine1","pointUpperBody")],
			("spine","hiResSpine3"):[("middleSpineIKControl","parent"),
										("middleSpineControl","parent")],
			("spine","hiResSpine5"):[("torsoIKControl","parent"),
										("torsoControl","parent")],
			("spine","hiResHips"):[("hipsIKControl","parentOffset"),
										("lowerSpineControl","parentOffset")],
			("spine","loResSpine1"):[("loResSpine1","pointUpperBody")],
			("spine","loResSpine2"):[("middleSpineIKControl","parent"),
										("middleSpineControl","parent")],
			("spine","loResSpine3"):[("torsoIKControl","parent"),
										("torsoControl","parent")],
			("spine","loResPelvis"):[("hipsIKControl","parentOffset"),
										("lowerSpineControl","parentOffset")],
			("leg","HiResFoot"):[("LegIKControl","footIK"),
										("FootControl","orient")],
			("leg","HiResToe"):[("ToeControl","orient"),
										("LegIKControl","toeRaiseSwivel")],
			("leg","HiResLowerLeg"):[("LowerLegControl","FKMidLeg"),
										("LegIKKneeControl","upperLeg"),
										("UpperLegControl","upperLeg")],
			("leg","LoResLowerLeg"):[("LowerLegControl","FKMidLeg"),
										("LegIKKneeControl","upperLeg"),
										("UpperLegControl","upperLeg")],
			("leg","LoResFoot"):[("LegIKControl","footIK"),
										("FootControl","orient")],
			("leg","LoResToe"):[("ToeControl","orient"),
										("LegIKControl","toeRaiseSwivel")],
			("arm","HiResClavical"):[("ClavicalControl","orient")],
			("arm","HiResHand"):[("HandControl","orient"),
										("ArmIKControl","parent")],
			("arm","HiResLowerArm"):[("LowerArmControl","FKMidArm"),
										("ArmIKElbowControl","upperArm"),
										("UpperArmControl","upperArm")],
			("arm","LoResClavical"):[("ClavicalControl","orient")],
			("arm","LoResLowerArm"):[("UpperArmControl","upperArm"),
										("ArmIKElbowControl","upperArm"),
										("LowerArmControl","FKMidArm")],
			("arm","LoResHand"):[("HandControl","orient"),
										("ArmIKControl","parent")],
			("finger","Joint1"):[("Control1","parentOffset")],
			("thumb","Joint1"):[("Control1","parentOffset")],
			("finger","Joint2"):[("Control2","parentOffset")],
			("thumb","Joint2"):[("Control2","parentOffset")],
			("finger","Joint3"):[("Control3","parentOffset")],
			("thumb","Joint3"):[("Control3","parentOffset")],
	}
	aimDict = {
		"LoResLowerArm":"LoResHand",
		"HiResLowerArm":"HiResHand",
		"LoResLowerLeg":"LoResFoot",
		"HiResLowerLeg":"HiResFoot",
	}
	hikDict = {
		"TSM3WorldJoint":"Reference",
		"spine1_hiResSpine1":"Hips",
		"left_HiResUpperLeg":"LeftUpLeg",
		"left_LoResUpperLeg":"LeftUpLeg",
		"left_HiResLowerLeg":"LeftLeg",
		"left_LoResLowerLeg":"LeftLeg",
		"left_HiResFoot":"LeftFoot",
		"right_HiResUpperLeg":"RightUpLeg",
		"right_HiResLowerLeg":"RightLeg",
		"right_LoResUpperLeg":"RightUpLeg",
		"right_LoResLowerLeg":"RightLeg",
		"right_HiResFoot":"RightFoot",
		"hiResSpine1":"Spine",
		"left_HiResUpperArm":"LeftArm",
		"left_HiResLowerArm":"LeftForeArm",
		"left_LoResUpperArm":"LeftArm",
		"left_LoResLowerArm":"LeftForeArm",
		"left_HiResHand":"LeftHand",
		"right_HiResUpperArm":"RightArm",
		"right_HiResLowerArm":"RightForeArm",
		"right_LoResUpperArm":"RightArm",
		"right_LoResLowerArm":"RightForeArm",
		"right_HiResHand":"RightHand",
		"head":"Head",
		"left_HiResToe":"LeftToeBase",
		"right_HiResToe":"RightToeBase",
		"left_HiResClavical":"LeftShoulder",
		"right_HiResClavical":"RightShoulder",
		"neck":"Neck",
		"hiResSpine2":"Spine1",
		"hiResSpine3":"Spine2",
		"hiResSpine4":"Spine3",
		"hiResSpine5":"Spine4",
		"hiResSpine6":"Spine5",
		"loResSpine2":"Spine1",
		"loResSpine3":"Spine2",
		"loResSpine4":"Spine3",
		"loResSpine5":"Spine4",
		"neck2":"Neck1",
		"neck3":"Neck2",
		"left_HiResUpperLegTwist":"LeafLeftUpLegRoll1",
		"left_HiResLowerLegTwist":"LeafLeftLegRoll1",
		"right_HiResUpperLegTwist":"LeafRightUpLegRoll1",
		"right_HiResLowerLegTwist":"LeafRightLegRoll1",
		"left_HiResUpperArmTwist":"LeafLeftArmRoll1",
		"left_HiResLowerArmTwist":"LeafLeftForeArmRoll1",
		"right_HiResUpperArmTwist":"LeafRightArmRoll1",
		"right_HiResLowerArmTwist":"LeafRightForeArmRoll1",
		"left_thumbJoint1":"LeftHandThumb1",
		"left_thumbJoint2":"LeftHandThumb2",
		"left_thumbJoint3":"LeftHandThumb3",
		"left_thumbJoint4":"LeftHandThumb4",
		"left_indexFingerMetacarpal":"LeftInHandIndex",
		"left_indexFingerJoint1":"LeftHandIndex1",
		"left_indexFingerJoint2":"LeftHandIndex2",
		"left_indexFingerJoint3":"LeftHandIndex3",
		"left_middleFingerMetacarpal":"LeftInHandMiddle",
		"left_middleFingerJoint1":"LeftHandMiddle1",
		"left_middleFingerJoint2":"LeftHandMiddle2",
		"left_middleFingerJoint3":"LeftHandMiddle3",
		"left_ringFingerMetacarpal":"LeftInHandRing",
		"left_ringFingerJoint1":"LeftHandRing1",
		"left_ringFingerJoint2":"LeftHandRing2",
		"left_ringFingerJoint3":"LeftHandRing3",
		"left_pinkyFingerMetacarpal":"LeftInHandPinky",
		"left_pinkyFingerJoint1":"LeftHandPinky1",
		"left_pinkyFingerJoint2":"LeftHandPinky2",
		"left_pinkyFingerJoint3":"LeftHandPinky3",
		"right_thumbJoint1":"RightHandThumb1",
		"right_thumbJoint2":"RightHandThumb2",
		"right_thumbJoint3":"RightHandThumb3",
		"right_thumbJoint4":"RightHandThumb4",
		"right_indexFingerMetacarpal":"RightInHandIndex",
		"right_indexFingerJoint1":"RightHandIndex1",
		"right_indexFingerJoint2":"RightHandIndex2",
		"right_indexFingerJoint3":"RightHandIndex3",
		"right_middleFingerMetacarpal":"RightInHandMiddle",
		"right_middleFingerJoint1":"RightHandMiddle1",
		"right_middleFingerJoint2":"RightHandMiddle2",
		"right_middleFingerJoint3":"RightHandMiddle3",
		"right_ringFingerMetacarpal":"RightInHandRing",
		"right_ringFingerJoint1":"RightHandRing1",
		"right_ringFingerJoint2":"RightHandRing2",
		"right_ringFingerJoint3":"RightHandRing3",
		"right_pinkyFingerMetacarpal":"RightInHandPinky",
		"right_pinkyFingerJoint1":"RightHandPinky1",
		"right_pinkyFingerJoint2":"RightHandPinky2",
		"right_pinkyFingerJoint3":"RightHandPinky3",	
	}
	def isTSM3Node(node):
		try:
			return node.hasAttr('TSM3NodeId')
		except KeyError:
			return False
	
	def nodeBodyPartId(node):
		""" Returns the body part ID for node"""
		try:
			return node.TSM3BodyPartId.get()
		except AttributeError:
			return None

	def nodeId(node):
		""" Returns the node ID for node"""
		try:
			return node.TSM3NodeId.get()
		except AttributeError:
			return None
		except KeyError:
			return None
	
	def nodeCompositeKey(node):
		""" Returns a nodes composite key, as a tuple, body part id first."""
		return nodeBodyPartId(node), nodeId(node)

	def nodeWithCompositeId(id1,id2):
		return nodeCache.get((id1,id2))

	def makeConstraint(source,target,type):
		"""makes the mocap constraint object based on the given type"""
		if type == "parent":
			return mocapParentConstraint(source,target)
		if type == "parentOffset":
			return mocapParentConstraint(source,target, maintainOffset = True)
		if type == "footIK":
			return mocapFootIKConstraint(source,target, mirror=isMirroredControl(target) and isNot2016())
		elif type == "upperLeg":
			return mocapUpperLimbConstraint(source,target, upZ = -1)
		elif type == "upperArm":
			return mocapUpperLimbConstraint(source,target, findAxis = True)
		elif type == "FKMidArm":
			return mocapFKMidLimbConstraint(source,target, findAxis = True) 
		elif type == "FKMidLeg":
			return mocapFKMidLimbConstraint(source, target)
		elif type == "orient":
			return mocapOrientConstraint(source,target)
		elif type == "pointUpperBody":
			return mocapParentConstraint(source,upperBodyFromRigNode(associatedRigNode(target)), maintainOffset = True)
		elif type == "toeRaiseSwivel":
			return mocapToeRaiseSwivelConstraint(source,target)

	def isMirroredControl(node):
		"""Returns true if world space scale Z is negative"""
		return pa.xform(node, q=True, scale=True, ws=True)[2]<0

	def isNot2016():
		return cmds.about(apiVersion=True) < 201600

	class mocapConstraint(object):
		def tryMocapConstraint(self):
			"""Tries to make the mocap constraint."""
			try:
				self.createAndIDConstraint()
			except Exception as e:
				om.MGlobal.displayError("Error creating mocap constraint for " + self.source + ". Rig may not follow Mocap skeleton as expected.")
		def createAndIDConstraint(self):
			"""creates and assigns an ID to the constraint"""
			self.assignIDConstraint(self.createConstraint())

		def createConstraint(self):
			"""throws an error. This should not be called in the base class."""
			om.MGlobal.displayError("Error creating mocap constraint. Mocap constraint has no type.")

		def assignIDConstraint(self, node):
			"""assigns a duplicate node ID"""
			node.addAttr("TSM3DupeNodeId", dt="string")
			node.TSM3DupeNodeId.set(self.target.TSM3NodeId.get() + "_MocapConstraint")

	class mocapUnitConversion(mocapConstraint):
		"""for fingers/thumbs/toes, attaches one attribute to another with a specific unit conversion factor""" 
		def __init__(self,sourceAttr,destAttr,factor):
			self.sourceAttr,self.destAttr, self.factor = sourceAttr, destAttr, factor

		def createConstraint(self):
			conversionName = self.destAttr.name().replace(".","_")+"_unitConversion"
			unitConversion = pa.createNode('unitConversion', name=conversionName)

			unitConversion.conversionFactor.set(self.factor)

			self.sourceAttr >> unitConversion.input
			unitConversion.output >> self.destAttr

			return unitConversion

		def assignIDConstraint(self, node):
			node.addAttr("TSM3DupeNodeId", dt = "string")
			node.TSM3DupeNodeId.set(node.name() + "_MocapConstraint")

	class mocapParentConstraint(mocapConstraint):
		"""basic parent constraint"""
		def __init__(self,source,target, maintainOffset = False):
			self.source, self.target, self.maintainOffset = source, target, maintainOffset

		def createConstraint(self):
			return pa.parentConstraint(self.source,self.target, mo = self.maintainOffset)

	class mocapFootIKConstraint(mocapConstraint):
		def __init__(self, source, target, mirror):
			self.source, self.target, self.mirror = source, target, mirror

		def createConstraint(self):
			constraint = mocapParentConstraint(self.source, self.target,maintainOffset = True).createConstraint()
			if self.mirror:
				pa.setAttr(constraint+".target[0].targetOffsetRotateY",180)

			return constraint
	class mocapPointConstraint(mocapConstraint):
		def __init__(self,source,target):
			self.source, self.target = source, target

		def createConstraint(self):
			return pa.pointConstraint(self.source,self.target, mo=True)

	class mocapOrientConstraint(mocapConstraint):
		def __init__(self,source,target):
			self.source, self.target = source, target

		def createConstraint(self):
			return pa.orientConstraint(self.source,self.target, mo=True)

	class mocapFKMidLimbConstraint(mocapConstraint):
		"""Mocap constraint for the middle joint in limbs (knee, elbow) in FK. In the case of arms
		must take into account the direction of the limb. In TSM3, this will also be the case for legs."""
		def __init__(self,source,target,findAxis = False):
			self.source,self.target = source, target		
			dupeBodyPartId = source.TSM3DupeBodyPartId.get()
			dupeNodeId =source.TSM3DupeNodeId.get()
			self.aimAt= dupeJointByCompositeId(dupeBodyPartId, aimDict[dupeNodeId])

		def createConstraint(self):
			return pa.aimConstraint(self.aimAt, self.target, aimVector = [1,0,0], upVector = [0,0,1], worldUpType="none",skip = ["x","z"])

	class mocapUpperLimbConstraint(mocapConstraint):
		"""For upper arm and upper leg, arms take into account the direction of the limb."""
		def __init__(self,source,target,findAxis = False, upZ=1):
			dupeBodyPartId = source.TSM3DupeBodyPartId.get()
			dupeNodeId =source.TSM3DupeNodeId.get()
			self.source,self.target = source, target
			self.upZ = upZ
			self.upObject= dupeJointByCompositeId(dupeBodyPartId, aimDict[dupeNodeId])

		def createConstraint(self):
			return pa.aimConstraint(self.source, self.target, aimVector = [1,0,0], upVector = [0,0,self.upZ], worldUpType="object", worldUpObject = self.upObject)

	class mocapToeRaiseSwivelConstraint(mocapConstraint):
		"""connects the toe raise and toe swivel attributes on the IK foot."""
		def __init__(self,source,target):
			self.source,self.target = source, target

		def createAndIDConstraint(self):
			#IK Leg Toe constraints do not require any additional naming, as that is part of unit conversion creation.
			sourceBodyPartId = self.source.TSM3DupeBodyPartId.get()

			mocapUnitConversion(self.source.rotateY,self.target.ToeRaise,-58.823).createAndIDConstraint()
			mocapUnitConversion(self.source.rotateZ,self.target.ToeSwivel,58.823).createAndIDConstraint()

	def rigNodeHasDupeSkeleton(rigNode):
		dupeJointList = [joint for joint in pa.listRelatives(rigNode, ad = True)
					if joint.hasAttr("TSM3DupeBodyPartId")]
		return len(dupeJointList) is not 0

	def verifySelection():
		"""Checks to make sure a TSM3 joint or control is found"""
		try:
			rigNode = associatedRigNode(pa.ls(sl=True)[0])
			if rigNodeHasDupeSkeleton(rigNode):
				raise Exception("Selection Invalid: Duplicate skeleton already exists for the selected TSM3 rig.")
		except IndexError:
			raise Exception("Selection invalid: To create the motion capture skeleton, please select part of the TSM3 rig.")

	def dupeJointByCompositeId(dupeBodyPartId,dupeNodeId):
		"""Using a dupe body part ID and dupe Node ID, return a dupe joint."""
		jointList = [joint for joint in pa.ls(type = "joint") if joint.hasAttr("TSM3DupeNodeId") and joint.TSM3DupeNodeId.get() == dupeNodeId 
				and joint.hasAttr("TSM3DupeBodyPartId") and joint.TSM3DupeBodyPartId.get() == dupeBodyPartId]
		if len(jointList) == 1:
			return jointList[0]
		else:
			return False

	def upperBodyFromRigNode(rigNode):
		"""Returns upper body node, given the "Rig" transform"""
		upperBodyList = [joint for joint in pa.listRelatives(rigNode, ad = True, typ = "transform")
			if joint.hasAttr("TSM3NodeId") and joint.TSM3NodeId.get() == "TSM3_upper_body"]
		if len(upperBodyList) == 1:
			return upperBodyList[0]
		else:
			return False

	def associatedRigNode(node):
		"""Find the "Rig" transform node based on a part of the rig.""" 
		return [parent for parent in node.getAllParents() if parent.hasAttr('TSM3NodeId') and parent.TSM3NodeId.get() == 'TSM3_Rig'][0]

	def worldJointFromRigNode(rigNode):
		"""Find world joint based on the "Rig" transform node"""
		worldJointList = [joint for joint in pa.listRelatives(rigNode, ad = True, typ = 'joint')
							if joint.hasAttr("TSM3NodeId") and joint.TSM3NodeId.get() == "TSM3WorldJoint"]
		if len(worldJointList) == 1:
			return worldJointList[0]
		else:
			return False

	def dupJointTSM3Atts(joint):
		"""removes TSM3 Node ID/Body Part ID and replaces it with TSM3 Dupe Node ID and TSM3 Dupe Body Part ID"""
		def replaceAttr(node, oldName, newName):
			if node.hasAttr(oldName):
				value = node.getAttr(oldName)
				node.setAttr(oldName,lock=False)
				node.deleteAttr(oldName)
				node.addAttr(newName, dt="string")
				node.setAttr(newName,value,lock=True)
		replaceAttr(joint,"TSM3NodeId","TSM3DupeNodeId")
		replaceAttr(joint,"TSM3BodyPartId","TSM3DupeBodyPartId")

		return joint
		

		
	def duplicateSkeleton(rootJoint):
		"""Create the duplicate skeleton with appropriate names and attributes, returns its joints"""
		controlSet = set([(control.TSM3BodyPartId.get(), control.TSM3Space.get()) for control in pa.ls(type='transform') if control.hasAttr('TSM3Space') == True and control.hasAttr('TSM3Control') == True])
		controlDict = {partId:handedness for partId, handedness in list(controlSet)}		

		
		def getHIKNames(origJoint):
			cleanName = origJoint.name()
			if origJoint.hasAttr('TSM3NodeId'):
				if origJoint.hasAttr('bodyPartType'):
					if origJoint.bodyPartType.get() == "finger" and "Finger1" in origJoint.name():
						cleanName = "pinkyFinger" + origJoint.TSM3NodeId.get()
					elif origJoint.bodyPartType.get() == "finger" and "Finger2" in origJoint.name():
						cleanName = "ringFinger" + origJoint.TSM3NodeId.get()
					elif origJoint.bodyPartType.get() == "finger" and "Finger3" in origJoint.name():
						cleanName = "middleFinger" + origJoint.TSM3NodeId.get()
					elif origJoint.bodyPartType.get() == "finger" and "Finger4" in origJoint.name():
						cleanName = "indexFinger"+ origJoint.TSM3NodeId.get()
					elif origJoint.bodyPartType.get() == "thumb":
						cleanName = "thumb" + origJoint.TSM3NodeId.get()
					else:
						cleanName = origJoint.TSM3NodeId.get()

					

			defaultName = cleanName
			defaultControlDict = 0
			if origJoint.hasAttr('TSM3BodyPartId') and origJoint.hasAttr('bodyPartType') and controlDict[origJoint.TSM3BodyPartId.get()] == 0 and origJoint.bodyPartType.get() == "spine":

				return hikDict.get(cleanName, defaultName)

			elif origJoint.hasAttr('TSM3BodyPartId') and origJoint.hasAttr('bodyPartType') and controlDict[origJoint.TSM3BodyPartId.get()] == 0 and origJoint.bodyPartType.get() == "head":

				return hikDict.get(cleanName, defaultName)
				
			elif origJoint.hasAttr('TSM3BodyPartId') and controlDict[origJoint.TSM3BodyPartId.get()] == 0:

				return hikDict.get("left_" + cleanName, defaultName)

			elif origJoint.hasAttr('TSM3BodyPartId') and controlDict[origJoint.TSM3BodyPartId.get()] == 1:

				return hikDict.get("right_" + cleanName, defaultName)
			
			else:
				return hikDict.get(cleanName, defaultName)
			return (cleanName)

				
		def renameDupJoint(origJoint,dupJoint):
			return dupJoint.rename(("dup_" + getHIKNames(origJoint)).replace(origJoint.namespace(),''))
			
		hierarchy = [rootJoint]
		hierarchy.extend(pa.listRelatives(rootJoint, ad=True, type = 'joint')) 

		return [dupJointTSM3Atts(renameDupJoint(origJoint,dupJoint)) for origJoint, dupJoint in zip(hierarchy, pa.duplicate(hierarchy, po = True, rc=True))]

	def mocapInfoFromJoint(joint):
		"""Returns a tuple with the joint, dupe body part ID, the key to the constraint dictionary (a tuple with the body part type and dupe node Id)"""
		def getAttrIfExists(attr):
			if joint.hasAttr(attr):
				return joint.getAttr(attr)
			else:
				return None

		return (joint, getAttrIfExists('TSM3DupeBodyPartId'),(getAttrIfExists('bodyPartType'),getAttrIfExists('TSM3DupeNodeId')))
	
	def worldJointFromTSM3Node(node):
		return worldJointFromRigNode(associatedRigNode(node))

	def createAndConstrainMocapSkeleton():
		"""Creates and constrains the mocap skeleton"""
		worldJoint = worldJointFromTSM3Node(pa.ls(sl=True)[0])
		dupSkeleton = duplicateSkeleton(worldJoint)
		pa.sets(dupSkeleton, name="mocapSkeletonJoints")

		try:
			for joint, bodyPartId, dictTuple in [mocapInfoFromJoint(joint) for joint in dupSkeleton]:
				relationships = mocapDict.get(dictTuple,[])
				[makeConstraint(joint,nodeWithCompositeId(bodyPartId,constrainedObject),constraintType).tryMocapConstraint() 
									for constrainedObject, constraintType in relationships]
		except:
			raise Exception("error occured while constraining skeleton.")
	def bodyPartTypeFromId(bodyPartId):
		"""Find the connect attributes to upper body node for a given body part ID, return its body part type"""
		attributesNode = nodeWithCompositeId(bodyPartId,"AttributesConnectToUpperBody")

		if attributesNode is None:
			om.MGlobal.displayError("Unable to identify type of body part with ID " + bodyPartId +". This could result in missing constraints.")
			return None

		if attributesNode.hasAttr("bodyPartType"):
			return attributesNode.bodyPartType.get()
		else:
			return None

	#Additional offsets need to be added to various constraints in 2014
	
	verifySelection()
	nodeCache = dict(filter(lambda ((x,y),z): x is not None, 
					[(nodeCompositeKey(node), node) for node in pa.ls() if isTSM3Node(node)]))
	bodyPartIds = set(ids[0] for ids in nodeCache.keys())
	createAndConstrainMocapSkeleton()